from dataclasses import dataclass
import typing

import pydantic

from . import _decoding

__all__ = [
    "extract_annotations",
    "generate_json_schema",
    "generate_function_json_schema",
    "_function",
    "procedure_model",
]


def extract_annotations(
    function: typing.Callable,
    custom_payload_annotation=...,
    custom_return_annotation=...,
):
    if custom_payload_annotation is ...:
        custom_payload_annotation = function.__annotations__.get("payload", ...)
    if custom_return_annotation is ...:
        custom_return_annotation = function.__annotations__.get("return", ...)
    return custom_payload_annotation, custom_return_annotation


def generate_json_schema(annotation):
    """
    Generates a JSON schema from an annotation.
    """
    if annotation is ...:
        return None

    model = pydantic.TypeAdapter(annotation)
    return model.json_schema()


def generate_function_json_schema(
    target: typing.Callable,
    custom_name: str = ...,
    custom_description: str | None = None,
    custom_payload_annotation=...,
    custom_return_annotation=...,
):
    """
    Generates a JSON schema from a function.
    """
    if custom_name is ...:
        custom_name = target.__name__
    if custom_description is None:
        custom_description = target.__doc__
    custom_payload_annotation, custom_return_annotation = extract_annotations(
        target, custom_payload_annotation, custom_return_annotation
    )
    payload_json_schema = generate_json_schema(custom_payload_annotation)
    return_json_schema = generate_json_schema(custom_return_annotation)
    return {
        "name": custom_name,
        "description": custom_description,
        "payload": payload_json_schema,
        "return": return_json_schema,
    }


class _function[I, O](typing.Protocol):
    __name__: str

    async def __call__(
        self,
        payload: I,
        *args,
        **kwargs,
    ) -> O: ...


@dataclass(kw_only=True, slots=True)
class procedure_model[I, O]:
    function: _function[I, O]
    name: str = ...
    description: str | None = None
    tags: set[str] = ...
    validate: bool = True
    payload_model: typing.Any = ...
    return_model: typing.Any = ...
    exceptions: set[type[Exception]] = ...
    json_schema: typing.Mapping = ...
    serialize_payload: typing.Callable[[bytes | str], I] = _decoding.serialize_any_json
    serialize_return: typing.Callable[[bytes | str], O] = _decoding.serialize_any_json

    def __post_init__(self):
        if not callable(self.function):
            raise ValueError("decorated function must be callable")
        self.payload_model, self.return_model = extract_annotations(
            self.function, self.payload_model, self.return_model
        )
        self.json_schema = generate_function_json_schema(
            self.function,
            self.name,
            self.description,
            custom_payload_annotation=self.payload_model,
            custom_return_annotation=self.return_model,
        )
        self.name = self.json_schema["name"]
        self.description = self.json_schema["description"]
        if self.validate:
            self.serialize_payload = _decoding.serialize_json(self.payload_model)
            self.serialize_return = _decoding.serialize_json(self.return_model)
        if self.exceptions is ...:
            self.exceptions = set()
        if self.tags is ...:
            self.tags = set()



import asyncio
import multiprocessing
import signal

from . import (
    _service,
    _session,
)

__all__ = [
    "serve_single",
    "serve_multiple",
]


def serve_single(
    client: _session.client_iface,
    service: _service.remote_service,
    *,
    stop_loop_on_exit: bool | None = None,
) -> None:
    if stop_loop_on_exit is None:
        stop_loop_on_exit = True

    session = _session.Almanet(client)

    async def begin() -> None:
        await session.join()
        await service._post_join_event.notify(session)

    async def end() -> None:
        await session.leave()
        if stop_loop_on_exit:
            loop.stop()

    loop = asyncio.get_event_loop()

    for s in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(s, lambda: loop.create_task(end()))

    loop.create_task(begin())
    if not loop.is_running():
        loop.run_forever()


def _initialize_new_process(
    client,
    service_uri: str,
    **kwargs,
) -> None:
    service = _service.get_service(service_uri)
    if service is None:
        raise ValueError(f"invalid service type {service_uri=}")

    serve_single(client, service, **kwargs)


def serve_multiple(
    sample_client: _session.client_iface,
    *services: _service.remote_service,
    **kwargs,
) -> None:
    if len(services) == 0:
        raise ValueError("must provide at least one service")

    if not all(isinstance(s, _service.remote_service) for s in services):
        raise TypeError("all services must be of type remote_service")

    processes: list[multiprocessing.Process] = []
    for s in services:
        client4process = sample_client.clone()
        process = multiprocessing.Process(
            target=_initialize_new_process,
            args=(client4process, s.pre),
            kwargs=kwargs,
        )
        process.start()
        processes.append(process)

    has_active_process = True
    while has_active_process:
        has_active_process = False

        for process in processes:
            if not process.is_alive():
                continue

            try:
                process.join()
            except KeyboardInterrupt:
                # termination signal to child processes already sent
                has_active_process = True



import asyncio
from datetime import datetime, timedelta, UTC
import logging
import typing

from . import _shared

__all__ = [
    "logger",
    "client_iface",
    "invoke_event_model",
    "qmessage_model",
    "reply_event_model",
    "rpc_exception",
    "Almanet",
    "get_active_session",
]


logger = logging.getLogger("almanet")

DEFAULT_CHANNEL = "almanet.python"


@_shared.dataclass(slots=True)
class qmessage_model[T: bytes]:
    """
    Represents a message in the queue.
    """

    id: str
    timestamp: int
    body: T
    attempts: int
    commit: typing.Callable[[], typing.Awaitable[None]]
    rollback: typing.Callable[[], typing.Awaitable[None]]

    @property
    def time(self) -> datetime:
        return datetime.fromtimestamp(self.timestamp, UTC)


type returns_consumer[T: bytes] = tuple[typing.AsyncIterable[qmessage_model[T]], typing.Callable[[], None]]


class client_iface(typing.Protocol):
    """
    Interface for a client library.
    """

    def clone(self) -> "client_iface":
        raise NotImplementedError()

    async def connect(
        self,
    ) -> None:
        raise NotImplementedError()

    async def produce(
        self,
        topic: str,
        message: str | bytes,
        delay: int,
    ) -> None:
        raise NotImplementedError()

    async def consume(
        self,
        topic: str,
        channel: str,
    ) -> returns_consumer[bytes]:
        raise NotImplementedError()

    async def close(self) -> None:
        raise NotImplementedError()


DEFAULT_TIMEOUT_SECONDS = 60


@_shared.dataclass(slots=True)
class invoke_event_model:
    """
    Represents an invocation event.
    """

    id: str
    caller_id: str
    payload: bytes
    reply_topic: str
    expires: int = DEFAULT_TIMEOUT_SECONDS


@_shared.dataclass(slots=True)
class reply_event_model:
    """
    Represents a reply event.
    """

    call_id: str
    is_exception: bool
    payload: bytes


@_shared.dataclass(slots=True)
class _reply_exception_model:
    name: str
    payload: bytes


class rpc_exception(Exception):
    __slots__ = ("name", "payload")

    def __init__(
        self,
        payload: typing.Any = None,
        name: str | None = None,
    ) -> None:
        self.name = name or self.__class__.__name__
        self.payload = payload

    def __str__(self) -> str:
        return f"{self.name}({self.payload})"

    __repr__ = __str__


@_shared.dataclass(slots=True)
class registration_model:
    """
    Represents a registered procedure to call.
    """

    uri: str
    channel: str
    procedure: typing.Callable
    session: "Almanet"

    @property
    def __name__(self):
        return self.uri

    async def execute(
        self,
        invocation: invoke_event_model,
    ) -> reply_event_model:
        __log_extra = {"registration": str(self), "invocation": str(invocation)}
        try:
            logger.debug(f"trying to execute procedure {self.uri}", extra=__log_extra)
            reply_payload = await self.procedure(
                invocation.payload,
                session=self.session,
            )
            is_exception = False
        except Exception as e:
            is_exception = True

            if isinstance(e, rpc_exception):
                reply_exception = _reply_exception_model(e.name, _shared.dump(e.payload))
            else:
                logger.exception("during execute procedure", extra=__log_extra)
                reply_exception = _reply_exception_model("InternalError", b"oops")

            reply_payload = _shared.dump(reply_exception)

        return reply_event_model(
            call_id=invocation.id,
            is_exception=is_exception,
            payload=reply_payload,
        )


class Almanet:
    """
    Represents a session, connected to message broker.
    """

    def __init__(
        self,
        client: client_iface,
    ) -> None:
        self.id = _shared.new_id()
        self.reply_topic = f"_rpc_._reply_.{self.id}#ephemeral"
        self.joined = False
        self._client = client
        self._background_tasks = _shared.background_tasks()
        self._post_join_event = _shared.observable_event()
        self._leave_event = _shared.observable_event()
        self._pending_replies: typing.MutableMapping[str, asyncio.Future[reply_event_model]] = {}

    @property
    def version(self) -> float:
        return 0.1

    async def _produce(
        self,
        uri: str,
        payload: typing.Any,
        delay: int = 0,
    ) -> None:
        try:
            message_body = _shared.dump(payload)
        except Exception as e:
            logger.error(f"during encode payload: {repr(e)}")
            raise e

        try:
            logger.debug(f"trying to produce {uri} topic")
            await self._client.produce(uri, message_body, delay=delay)
        except Exception as e:
            logger.exception(f"during produce {uri} topic")
            raise e

    def produce(
        self,
        uri: str,
        payload: typing.Any,
        delay: int = 0,
    ) -> asyncio.Task[None]:
        """
        Produce a message with a specified topic and payload.
        """
        return self._background_tasks.schedule(self._produce(uri, payload, delay=delay))

    async def consume(
        self,
        topic: str,
        channel: str,
    ) -> returns_consumer:
        """
        Consume messages from a message broker with the specified topic and channel.
        It returns a tuple of a stream of messages and a function that can stop consumer.
        """
        logger.debug(f"trying to consume {topic}:{channel}")

        messages_stream, stop_consumer = await self._client.consume(topic, channel)

        def __stop_consumer():
            logger.warning(f"trying to stop consumer {topic}")
            stop_consumer()

        self._leave_event.add_observer(__stop_consumer)

        return messages_stream, __stop_consumer

    async def _consume_replies(
        self,
        ready_event: asyncio.Event,
    ) -> None:
        messages_stream, _ = await self.consume(
            self.reply_topic,
            channel=f"{DEFAULT_CHANNEL}#ephemeral",
        )
        logger.debug("reply event consumer begin")
        serializer = _shared.serialize_json(reply_event_model)
        ready_event.set()
        async for message in messages_stream:
            __log_extra = {"incoming_message": str(message)}
            try:
                reply_event = serializer(message.body)
                __log_extra["reply_event"] = str(reply_event)
                logger.debug("new reply", extra=__log_extra)

                pending = self._pending_replies.get(reply_event.call_id)
                if pending is None:
                    logger.warning("pending event not found", extra=__log_extra)
                else:
                    pending.set_result(reply_event)
            except:
                logger.exception("during parse reply", extra=__log_extra)

            await message.commit()
            logger.debug("successful commit", extra=__log_extra)
        logger.debug("reply event consumer end")

    _call_args = tuple[str, typing.Any]

    class _call_kwargs(typing.TypedDict):
        timeout: typing.NotRequired[int]

    async def _delay_call(
        self,
        uri: str,
        payload: typing.Any,
        delay: int = 0,
        /,
        _invocation_id: str | None = None,
        _reply_topic: str = "",
        _expires: int = 60,
    ) -> None:
        invocation = invoke_event_model(
            id=_invocation_id or _shared.new_id(),
            caller_id=self.id,
            payload=_shared.dump(payload),
            reply_topic=_reply_topic,
            expires=_expires,
        )

        __log_extra = {"invoke_event": str(invocation)}
        logger.debug(f"trying to call {uri=} {delay=}", extra=__log_extra)

        await self._produce(f"_rpc_.{uri}", invocation, delay=delay)

    def delay_call(
        self,
        uri: str,
        payload,
        delay: int = 0, # seconds
    ) -> asyncio.Task[None]:
        return self._background_tasks.schedule(self._delay_call(uri, payload, delay))

    async def _call(
        self,
        uri: str,
        payload,
        timeout: int = DEFAULT_TIMEOUT_SECONDS,
    ) -> reply_event_model:
        invocation_id = _shared.new_id()

        __log_extra = {"uri": uri, "timeout": timeout, "invocation_id": invocation_id}

        try:
            async with asyncio.timeout(timeout):
                pending_reply_event = asyncio.Future[reply_event_model]()
                self._pending_replies[invocation_id] = pending_reply_event

                await self._delay_call(
                    uri,
                    payload,
                    _invocation_id=invocation_id,
                    _reply_topic=self.reply_topic,
                    _expires=timeout,
                )

                reply_event = await pending_reply_event
                __log_extra["reply_event"] = str(reply_event)
                logger.debug(f"invocation {uri=} respond", extra=__log_extra)

                if reply_event.is_exception:
                    serializer = _shared.serialize_json(_reply_exception_model)
                    reply_exception = serializer(reply_event.payload)
                    raise rpc_exception(
                        reply_exception.payload,
                        name=reply_exception.name,
                    )

                return reply_event
        except Exception as e:
            logger.error(f"during call {uri}: {e!r}", extra={**__log_extra, "error": str(e)})
            raise e
        finally:
            self._pending_replies.pop(invocation_id)

    def call(
        self,
        uri: str,
        payload,
        **kwargs: typing.Unpack[_call_kwargs],
    ) -> asyncio.Task[reply_event_model]:
        """
        Executes the remote procedure using the payload.
        Returns a instance of result model.
        """
        return self._background_tasks.schedule(self._call(uri, payload, **kwargs))

    async def _multicall(
        self,
        uri: str,
        payload,
        timeout: int = DEFAULT_TIMEOUT_SECONDS,
    ) -> list[reply_event_model]:
        reply_topic = f"_rpc_._replies_.{_shared.new_id()}#ephemeral"

        __log_extra = {"uri": uri, "timeout": timeout, "reply_topic": reply_topic}

        messages_stream, stop_consumer = await self.consume(
            reply_topic,
            f"{DEFAULT_CHANNEL}#ephemeral"
        )

        serializer = _shared.serialize_json(reply_event_model)

        result = []
        try:
            async with asyncio.timeout(timeout):
                await self._delay_call(
                    uri,
                    payload,
                    _reply_topic=reply_topic,
                )

                async for message in messages_stream:
                    try:
                        logger.debug("new reply event", extra=__log_extra)
                        reply_event = serializer(message.body)
                        result.append(reply_event)
                    except:
                        logger.exception("during parse reply event", extra=__log_extra)

                    await message.commit()
        except TimeoutError:
            stop_consumer()

        logger.debug(f"multicall {uri} done")

        return result

    def multicall(
        self,
        uri: str,
        payload,
        **kwargs: typing.Unpack[_call_kwargs],
    ) -> asyncio.Task[list[reply_event_model]]:
        """
        Execute simultaneously multiple procedures using the payload.
        """
        return self._background_tasks.schedule(self._multicall(uri, payload, **kwargs))

    async def __execution(
        self,
        registration: registration_model,
        message: qmessage_model[bytes],
    ):
        __log_extra = {"registration": str(registration), "incoming_message": str(message)}
        serializer = _shared.serialize_json(invoke_event_model)
        try:
            invocation = serializer(message.body)
            __log_extra["invocation"] = str(invocation)

            expiration_time = message.time + timedelta(seconds=invocation.expires)

            def check_expiration():
                current_time = datetime.now(UTC)
                delta = expiration_time - current_time
                if delta.total_seconds() <= 0:
                    raise TimeoutError(
                        f"invocation expired after execution! delay={delta}s"
                    )
                return delta.total_seconds()

            logger.debug(f"new invocation {expiration_time=}", extra=__log_extra)

            remaining_seconds = check_expiration()

            async with asyncio.timeout(remaining_seconds):
                reply_event = await registration.execute(invocation)

            check_expiration()

            if len(invocation.reply_topic) > 0:
                logger.debug(f"trying to reply {registration.uri}", extra=__log_extra)
                await self._produce(invocation.reply_topic, reply_event)
        except Exception as e:
            logger.error(f"during execute invocation {e!r}", extra=__log_extra)
        finally:
            await message.commit()
            logger.debug("successful commit", extra=__log_extra)

    async def _consume_invocations(
        self,
        registration: registration_model,
    ) -> None:
        logger.debug(f"trying to register {registration.uri}:{registration.channel}")
        messages_stream, _ = await self.consume(f"_rpc_.{registration.uri}", registration.channel)
        async for message in messages_stream:
            self._background_tasks.schedule(self.__execution(registration, message))
            if not self.joined:
                break
        logger.debug(f"consumer {registration.uri} down")

    def register(
        self,
        topic: str,
        procedure: typing.Callable,
        *,
        channel: str = "main",
    ) -> registration_model:
        """
        Register a procedure with a specified topic and payload.
        Returns the created registration.
        """
        if not self.joined:
            raise RuntimeError(f"session {self.id} not joined")

        logger.debug(f"scheduling registration {topic}")

        registration = registration_model(
            uri=topic,
            channel=channel,
            procedure=procedure,
            session=self,
        )

        self._background_tasks.schedule(self._consume_invocations(registration), daemon=True)

        return registration

    async def join(
        self,
    ) -> None:
        """
        Join the session to message broker.
        """
        if self.joined:
            raise RuntimeError(f"session {self.id} already joined")

        logger.debug("trying to connect")

        await self._client.connect()

        consume_replies_ready = asyncio.Event()
        self._background_tasks.schedule(
            self._consume_replies(consume_replies_ready),
            daemon=True,
        )
        await consume_replies_ready.wait()

        _active_session.set(self)

        self.joined = True
        await self._post_join_event.notify()
        logger.info(f"session {self.id} joined")

    async def __aenter__(self) -> "Almanet":
        if not self.joined:
            await self.join()
        return self

    async def leave(
        self,
        reason: str | None = None,
        timeout: float = 60,
    ) -> None:
        """
        Leave the session from message broker.
        """
        if not self.joined:
            raise RuntimeError(f"session {self.id} not joined")

        _active_session.set(None)

        self.joined = False

        logger.debug(f"trying to leave {self.id} session, reason: {reason}")

        logger.debug(f"session {self.id} await task pool complete")
        await self._background_tasks.complete(timeout=timeout)

        await self._leave_event.notify(timeout=timeout)

        logger.debug(f"session {self.id} trying to close connection")
        await self._client.close()

        logger.warning(f"session {self.id} left")

    async def __aexit__(self, etype, evalue, etraceback) -> None:
        if self.joined:
            await self.leave()


_active_session = _shared.new_concurrent_context()


def get_active_session() -> Almanet:
    session = _active_session.get(None)
    if session is None:
        raise RuntimeError("active session not found")
    return session



import typing

import pydantic_core

from . import _session
from . import _shared

__all__ = [
    "remote_exception",
    "rpc_invalid_payload",
    "rpc_invalid_return",
    "remote_procedure_model",
    "remote_service",
    "new_remote_service",
]


class remote_exception(_session.rpc_exception):
    """
    Represents an remote procedure call exception.
    You can inherit from this class to create your own exceptions.
    """

    payload: typing.Any

    @classmethod
    def _make_from_payload(
        klass,
        v: bytes,
        *args,
        **kwargs,
    ) -> "remote_exception":
        """
        Returns instance of the class with the serialized payload.

        Raises:
        - pydantic_code.ValidationError
        """
        model = klass.__annotations__.get("payload", ...)
        serializer = _shared.serialize_json(model)
        payload = serializer(v)
        return klass(payload)


class rpc_invalid_payload(remote_exception):
    """
    Represents an invalid payload for a remote procedure call.
    """

    payload: typing.Any


class rpc_invalid_return(remote_exception):
    """
    Represents an invalid return value for a remote procedure call.
    """


class rpc_invalid_exception_payload(rpc_invalid_payload):
    """
    Represents an invalid payload for remote exception.
    """


@_shared.dataclass(kw_only=True, slots=True)
class remote_procedure_model[I, O](_shared.procedure_model[I, O]):
    service: "remote_service"
    uri: str = ...
    channel: str = _session.DEFAULT_CHANNEL
    exceptions: set[type[remote_exception]] = ...
    include_to_api: bool = False
    _has_implementation: bool = False

    def __post_init__(self):
        super(remote_procedure_model, self).__post_init__()
        if self.uri is ...:
            self.uri = ".".join([self.service.pre, self.name])
        self.exceptions.add(rpc_invalid_payload)
        self.exceptions.add(rpc_invalid_return)

    def execute(
        self,
        payload: I,
        session: _session.Almanet,
    ) -> typing.Awaitable[O]:
        return self.function(payload, session=session)

    async def _remote_execution(
        self,
        payload: bytes,
        session: _session.Almanet,
    ) -> bytes:
        """
        if called remotely
        """
        _session.logger.debug(f"remote calling {self.uri}")

        try:
            __payload = self.serialize_payload(payload)
        except pydantic_core.ValidationError as e:
            raise rpc_invalid_payload(str(e))

        result = await self.execute(__payload, session)

        # if not isinstance(result, self.return_model):
        #     raise rpc_invalid_return()

        return _shared.dump(result)

    class _local_execution_kwargs(_session.Almanet._call_kwargs):
        force_local: typing.NotRequired[bool]

    async def _local_execution(
        self,
        payload: I,
        **kwargs: typing.Unpack[_local_execution_kwargs],
    ) -> O:
        """
        if called locally

        Args:
        - force_local: if True, force local execution
        """
        _session.logger.debug(f"local execution {self.uri}")

        session = _session.get_active_session()

        force_local = kwargs.pop("force_local", True)
        if self._has_implementation and force_local:
            _session.logger.debug(f"local calling {self.uri}")
            return await self.execute(payload, session)

        try:
            reply_event = await session.call(self.uri, payload, **kwargs)
            return self.serialize_return(reply_event.payload)
        except _session.rpc_exception as e:
            for etype in self.exceptions:
                if e.name == etype.__name__:
                    try:
                        raise etype._make_from_payload(e.payload)
                    except pydantic_core.ValidationError as e:
                        raise rpc_invalid_exception_payload(str(e))
            _session.logger.warning(f"{e.name} exception not define for {self.uri}")
            raise e

    def __call__(
        self,
        payload: I,
        **kwargs: typing.Unpack[_local_execution_kwargs],
    ) -> typing.Awaitable[O]:
        return self._local_execution(payload, **kwargs)

    def implements(
        self,
        real_function: _shared._function[I, O],
    ) -> "remote_procedure_model[I, O]":
        if self._has_implementation:
            raise ValueError("procedure already implemented")

        procedure = self.service.add_procedure(
            real_function,
            name=self.name,
            include_to_api=self.include_to_api,
            description=self.description,
            tags=self.tags,
            validate=self.validate,
            payload_model=self.payload_model,
            return_model=self.return_model,
            uri=self.uri,
            channel=self.channel,
            exceptions=self.exceptions,
        )
        self.function = procedure.function

        self._has_implementation = True

        return procedure


class remote_service:
    def __init__(
        self,
        prepath: str,
        tags: set[str] | None = None,
        include_to_api: bool = False,
    ) -> None:
        self.pre: str = prepath
        self.default_tags: set[str] = set(tags or [])
        self.include_to_api: bool = include_to_api
        self.procedures: list[remote_procedure_model] = []
        self.background_tasks = _shared.background_tasks()
        self._post_join_event = _shared.observable_event()
        self._post_join_event.add_observer(self._share_all)
        _registry[self.pre] = self

    @property
    def routes(self) -> set[str]:
        return {f"{i.uri}:{i.channel}" for i in self.procedures}

    def post_join[T: typing.Callable](
        self,
        function: T,
    ) -> T:
        self._post_join_event.add_observer(function)
        return function

    class _register_procedure_kwargs(typing.TypedDict):
        name: typing.NotRequired[str]
        include_to_api: typing.NotRequired[bool]
        description: typing.NotRequired[str | None]
        tags: typing.NotRequired[set[str]]
        validate: typing.NotRequired[bool]
        payload_model: typing.NotRequired[typing.Any]
        return_model: typing.NotRequired[typing.Any]
        uri: typing.NotRequired[str]
        channel: typing.NotRequired[str]
        exceptions: typing.NotRequired[set[type[remote_exception]]]

    @typing.overload
    def public_procedure[I, O](
        self,
        function: _shared._function[I, O],
    ) -> remote_procedure_model[I, O]: ...

    @typing.overload
    def public_procedure[I, O](
        self,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> typing.Callable[[_shared._function[I, O]], remote_procedure_model[I, O]]: ...

    def public_procedure(
        self,
        function=None,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> remote_procedure_model | typing.Callable[[_shared._function], remote_procedure_model]:
        if function is None:
            return lambda function: remote_procedure_model(service=self, function=function, **kwargs)
        return remote_procedure_model(service=self, function=function, **kwargs)

    def add_procedure(
        self,
        function: typing.Callable,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> remote_procedure_model:
        procedure = remote_procedure_model(
            **kwargs,
            function=function,
            service=self,
            _has_implementation=True,
        )
        self.procedures.append(procedure)
        return procedure

    @typing.overload
    def procedure[I, O](
        self,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> typing.Callable[[_shared._function[I, O]], remote_procedure_model[I, O]]: ...

    @typing.overload
    def procedure[I, O](
        self,
        function: _shared._function[I, O],
    ) -> remote_procedure_model[I, O]: ...

    def procedure(
        self,
        function=None,
        **kwargs: typing.Unpack[_register_procedure_kwargs],
    ) -> remote_procedure_model | typing.Callable[[_shared._function], remote_procedure_model]:
        """
        Allows you to easily add procedures (functions) to a microservice by using a decorator.
        Returns a decorated function.
        """
        if function is None:
            return lambda function: self.add_procedure(function, **kwargs)
        return self.add_procedure(function, **kwargs)

    def _share_self_schema(
        self,
        session: _session.Almanet,
        **extra,
    ) -> None:
        async def procedure(*args, **kwargs):
            return {
                "session_id": session.id,
                "session_version": session.version,
                "routes": list(self.routes),
                **extra,
            }

        session.register(
            "_schema_.client",
            procedure,
            channel=session.id,
        )

    def _share_procedure_schema(
        self,
        session: _session.Almanet,
        registration: remote_procedure_model,
    ) -> None:
        tags = registration.tags | self.default_tags
        if len(tags) == 0:
            tags = {"default"}

        async def procedure(*args, **kwargs):
            return {
                "session_id": session.id,
                "session_version": session.version,
                "uri": registration.uri,
                "channel": registration.channel,
                "validate": registration.validate,
                "tags": tags,
                **registration.json_schema,
            }

        session.register(
            f"_schema_.{registration.uri}.{registration.channel}",
            procedure,
            channel=registration.channel,
        )

    def _share_all(
        self,
        session: _session.Almanet,
    ) -> None:
        _session.logger.info(f"Sharing {self.pre} procedures")

        for procedure in self.procedures:
            session.register(
                procedure.uri,
                procedure._remote_execution,
                channel=procedure.channel,
            )

            if procedure.include_to_api:
                self._share_procedure_schema(session, procedure)

        if self.include_to_api:
            self._share_self_schema(session)


new_remote_service = remote_service

_registry = {}


def get_service(
    uri: str,
) -> remote_service | None:
    """
    Returns the service object for the given uri.
    """
    return _registry.get(uri)



import asyncio
from . import _background_tasks

__all__ = ["observable_event"]


class observable_event:
    def __init__(self):
        self.task_pool = _background_tasks.background_tasks()
        self.observers = []

    def add_observer(self, function):
        self.observers.append(function)

    def observer(self, function):
        self.add_observer(function)
        return function

    def notify(self, *args, timeout: float | None = None, **kwargs):
        for observer in self.observers:
            result = observer(*args, **kwargs)
            if asyncio.iscoroutine(result):
                self.task_pool.schedule(result)

        return self.task_pool.complete(timeout)



import typing

import ansq

from almanet import _session
from almanet import _shared

if typing.TYPE_CHECKING:
    from ansq.tcp.types import NSQMessage

__all__ = [
    "ansqd_tcp_client",
    "make_ansqd_tcp_session",
]


class ansqd_tcp_client:
    def __init__(
        self,
        *addresses: str,
        message_timeout_seconds: float = 60,
    ):
        if len(addresses) == 0:
            raise ValueError("at least one address must be specified")

        if not all(isinstance(i, str) for i in addresses):
            raise ValueError("addresses must be a iterable of strings")

        self.addresses = addresses

        try:
            self.message_timeout_seconds = float(message_timeout_seconds)
        except Exception:
            raise ValueError("`message_timeout_seconds` must be float")

    def clone(self) -> "ansqd_tcp_client":
        return ansqd_tcp_client(*self.addresses)

    async def connect(self) -> None:
        self.writer = await ansq.create_writer(
            nsqd_tcp_addresses=self.addresses,
        )

    async def close(self) -> None:
        await self.writer.close()

    async def produce(
        self,
        topic: str,
        message: str | bytes,
        delay: int,
    ) -> None:
        if delay > 0:
            await self.writer.dpub(topic, message, delay)
        else:
            await self.writer.pub(topic, message)

    async def _convert_ansq_message(
        self,
        ansq_messages_stream: typing.AsyncIterable["NSQMessage"],
    ) -> typing.AsyncIterable[_session.qmessage_model[bytes]]:
        async for ansq_message in ansq_messages_stream:
            almanet_message = _session.qmessage_model(
                id=ansq_message.id,
                timestamp=ansq_message.timestamp / 1_000_000_000,
                body=ansq_message.body,
                attempts=ansq_message.attempts,
                commit=ansq_message.fin,
                rollback=ansq_message.req,
            )
            yield almanet_message

    async def consume(
        self,
        topic: str,
        channel: str,
    ) -> _session.returns_consumer:
        reader = await ansq.create_reader(
            nsqd_tcp_addresses=self.addresses,
            topic=topic,
            channel=channel,
            connection_options=ansq.ConnectionOptions(
                features=ansq.ConnectionFeatures(
                    msg_timeout=round(self.message_timeout_seconds * 1000)
                )
            ),
        )
        messages_stream = reader.messages()
        messages_stream = self._convert_ansq_message(messages_stream)
        # ansq does not close stream automatically
        return _shared.make_closable(messages_stream, reader.close)


def make_ansqd_tcp_session(*addresses: str) -> _session.Almanet:
    client = ansqd_tcp_client(*addresses)
    return _session.Almanet(client)
